#include "StdAfx.h"
#ifdef DOCON_SETUP
#include "TbSetup.h"
#else
#include "Com.h"
#endif

CCom::CCom()
{
	SetDefault();
}

CCom::~CCom()
{
	ComClose();
}

CCom::CCom(int typ, char *szComPort,int ComValue, unsigned short Absender)
{
	SetDefault();
	switch(typ) {
	case T_S2_RS232: // RS232 direkt
	case T_S2_BIM_DIREKT: // Funk mit BIM2 ber Umsetzer ohne eigenen Controller
		strcpy((char*)ComPort, szComPort);
		Baud = ComValue;
		break;
	case T_S2_TCP_IP: // TCP/IP
		if (strlen(szComPort) < MAX_REMOTE_IP) {
			strcpy(RemoteIP,szComPort);
		}
		PortNo = ComValue;
		break;
	}
	AdAbsender = Absender;
	Typ = typ;
}

void CCom::SetDefault()
// Grundeinstellungen 
{
	ComOpenFlag = FALSE;
	strcpy((char*)ComPort, "");
	Baud       = 38400;
	ComTimeOut = 500;
	hcom       = INVALID_HANDLE_VALUE;

	strcpy(RemoteIP,"");
	PortNo = 0;
	Socket = INVALID_SOCKET;
	
	CBFunction = NULL;
	AdAbsender = 0;
	PaketNr    = 0;
	TimeToLive = P_S2_TIME_TO_LIVE;
	
	LogIsOn = false;
	strcpy(LogDateiName, "Default.log");
}

void CCom::ComLog(const char *Format, ...)
{
	va_list ap;
	FILE *Logfp;

	if (LogIsOn) {
		va_start(ap, Format);

		if ((Logfp = fopen(LogDateiName, "a")) == (FILE *)0) {

			return;
		}
		vfprintf(Logfp, Format, ap);

		va_end(ap);
		fclose(Logfp);
	}
}

int CCom::ComOpen()
{
	int Ret;

	if (ComOpenFlag) return TRUE;
	if (Typ == T_S2_RS232 || Typ == T_S2_BIM_DIREKT) {
		Ret = SerialOpen((char*)ComPort,Baud,ComTimeOut);
	}else if (Typ == T_S2_TCP_IP) {
		Ret = SocketOpen(RemoteIP,PortNo,ComTimeOut);
	}else{
		Ret = FALSE;
	}
	if (Ret) {
		ComOpenFlag = TRUE; 
		return TRUE;
	}else{
		ComOpenFlag = FALSE;
		return FALSE;
	}
}

void CCom::ComClose()
{
	if (ComOpenFlag) {
		if (Typ == T_S2_RS232 || Typ == T_S2_BIM_DIREKT) {
			SerialClose();
		}else{
			SocketClose();
		}
		ComOpenFlag = FALSE;
	}
}

int CCom::SerialOpen(char *szComPort, int iBaudRate, int iTimeOut)
{
	int rc;

	ComLog("SerialOpen.Begin\n");

	ComTimeOut = iTimeOut;

#ifdef _WIN32

	if ((hcom = CreateFile(
				szComPort,						// pointer to name of the file 
				(GENERIC_READ | GENERIC_WRITE),	// access (read-write) mode 
				0,								// share mode 
				NULL,							// pointer to security descriptor 
				OPEN_EXISTING,					// how to create 
				0,								// file attributes 
				0 								// handle to file with attributes to copy  
				)) == INVALID_HANDLE_VALUE) { 

		return FALSE;
	}
#else
	errno = 0;

	if ((rc = open(szComPort, O_RDWR|O_NDELAY)) < 0) {
		ComLog("SerialOpen.open: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}

	hcom = rc;

	if ((rc = fcntl(hcom, F_SETFL, O_RDWR)) < 0) {
		ComLog("SerialOpen.fcntl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}

	flock_s.l_type    = F_WRLCK;  // lock stream for read and write
	flock_s.l_whence  = 0;        // lock whole file -> 0,0,0
	flock_s.l_start   = 0;
	flock_s.l_len     = 0;

	if ((rc = fcntl(hcom, F_SETLK, &flock_s)) < 0) {
		ComLog("SerialOpen.fcntl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}
#endif

	rc = SetSerialComState(hcom, iBaudRate, iTimeOut);

	ComLog("SerialOpen.End\n");

	return rc;
}

#ifdef _WIN32
int CCom::SetComTimeout(int TimeOutMS) 
{
	static COMMTIMEOUTS	ctout;
	static LPCOMMTIMEOUTS	pctout = &ctout;

	if (!GetCommTimeouts(hcom,pctout)) {
		CloseHandle(hcom);return 0;
	}
	ctout.ReadTotalTimeoutConstant = (DWORD)TimeOutMS;
	ctout.WriteTotalTimeoutConstant = (DWORD)100;
	ctout.WriteTotalTimeoutMultiplier = (DWORD)4;
	if (!SetCommTimeouts(hcom,pctout)) {
		CloseHandle(hcom);return 0;
	}
	return TRUE;
}
#endif

int CCom::SetSerialComState(HANDLE hComPort, int iBaudRate, int iTimeOut)
{
#ifdef _WIN32
	DCB          dcb;
	COMMTIMEOUTS ctout;
	DWORD        cerr = 0;

	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	dcb.fBinary   = 1;

	dcb.BaudRate  = iBaudRate;
	dcb.ByteSize  = 8;
	dcb.Parity    = NOPARITY;
	dcb.StopBits  = ONESTOPBIT;


	if (!SetCommState(hComPort, &dcb)) {

		CloseHandle(hComPort);
		return FALSE;
	}

	ctout.ReadIntervalTimeout         = 0;
	ctout.ReadTotalTimeoutMultiplier  = 0;
	ctout.ReadTotalTimeoutConstant    = (DWORD) iTimeOut;
	ctout.WriteTotalTimeoutConstant   = 0;
	ctout.WriteTotalTimeoutMultiplier = 0;

	if (!SetCommTimeouts(hComPort, &ctout)) {

		CloseHandle(hComPort);
		return FALSE;
	}

	// unsolved: fuer modems + RS485 notwendig !?!
	if (!EscapeCommFunction(hComPort, SETRTS)) {

		CloseHandle(hComPort);
		return FALSE;
	}
	if (!EscapeCommFunction(hComPort, SETDTR)) {

		CloseHandle(hComPort);
		return FALSE;
	}

	if (!ClearCommError(hComPort, &cerr, NULL)) {

		CloseHandle(hComPort);
		return FALSE;
	}

#else
	int rc;
	int baudmode, charmode, paricheck, stopmode;
	int termio_cm;

	switch(iBaudRate) {
	case 9600:
		baudmode = B9600;
		break;
	case 19200:
		baudmode = B19200;
	case 38400:
		baudmode = B38400;
		break;
	default:
		return FALSE;
	}

	charmode = CS8;
	paricheck = 0000000;
	stopmode = 0000000;

	termio_s.c_iflag = 0;
	termio_s.c_oflag = 0;
	termio_s.c_cflag = 0;
	termio_s.c_lflag = 0;
	termio_s.c_line  = 0;
	(void) memset((void *)termio_s.c_cc, 0, NCC);
   
	termio_s.c_cflag |= HUPCL;        // hangup line when close/exit(DTR of)
	termio_s.c_cflag |= CLOCAL;       // local/direct connection
   
	termio_s.c_cflag |= baudmode;     // set baudrate
	termio_s.c_cflag |= charmode;     // set databits
	termio_s.c_cflag |= paricheck;    // set parity
	termio_s.c_cflag |= stopmode;     // set stopbits

	termio_s.c_cflag |= CREAD;        // enable input queue

	termio_s.c_cc[VMIN]  = 1;         // read waiting for 1 char
   
	if ((rc = ioctl(hComPort, TCSETA, &termio_s)) < 0) {
		ComLog("SetSerialComState.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}

	if ((rc = ioctl(hComPort, TCFLSH, TCIOFLUSH)) < 0) {
		ComLog("SetSerialComState.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}

	if ((rc = ioctl(hComPort, TIOCMGET, &termio_cm)) < 0) {
		ComLog("SetSerialComState.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}

	if (! (termio_cm & TIOCM_RTS)) termio_cm |= TIOCM_RTS;
	if (! (termio_cm & TIOCM_DTR)) termio_cm |= TIOCM_DTR;

	if ((rc = ioctl(hComPort, TIOCMSET, &termio_cm)) < 0) {
		ComLog("SetSerialComState.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}
#endif

	return TRUE;
}

void CCom::SerialClose(void)
{
	ComLog("SerialClose.Begin\n");

#ifdef _WIN32
	EscapeCommFunction(hcom,CLRRTS);
	EscapeCommFunction(hcom,CLRDTR);
	CloseHandle(hcom);
#else
	int rc;
	int termio_cm;

	if ((rc = ioctl(hcom, TIOCMGET, &termio_cm)) < 0) {
		ComLog("SerialClose.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
	}

	if ((termio_cm & TIOCM_RTS)) termio_cm &= ~TIOCM_RTS;
	if ((termio_cm & TIOCM_DTR)) termio_cm &= ~TIOCM_DTR;

	if ((rc = ioctl(hcom, TIOCMSET, &termio_cm)) < 0) {
		ComLog("SerialClose.ioctl: failed (rc=%d) (errno=%d)\n", rc, errno);
	}

	close(hcom);
#endif

	hcom = INVALID_HANDLE_VALUE;

	ComLog("SerialClose.End\n");
}

int CCom::SocketOpen(char *RemoteIP, int PortNo, long TimeOut)
{
	SOCKADDR_IN sin;
	memset(&sin, 0, sizeof(sin));
   	sin.sin_family = AF_INET;
	sin.sin_port = htons(PortNo);
	sin.sin_addr.s_addr = inet_addr(RemoteIP);

	ComTimeOut  = TimeOut;

	ComLog("SocketOpen.Begin: (%s:%d)\n", RemoteIP, PortNo);

	if (Socket != INVALID_SOCKET) {
		ComLog("SocketOpen: Socket schon in Verwendung.\n");
		return FALSE;
	}

	if (sin.sin_addr.s_addr == INADDR_NONE)	{ // Adresse ber DNS auflsen
		HOSTENT *he;
		he = gethostbyname(RemoteIP);
		if (! he) {
			ComLog("SocketOpen.gethostbyname: failed(%d)\n", WSAGetLastError());
			return FALSE;
		}
		sin.sin_addr.s_addr = *((unsigned long*)he->h_addr);
	}

	if ((Socket = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
		ComLog("SocketOpen.socket: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}
      
	if (connect(Socket,(SOCKADDR*)&sin,sizeof(sin)) == SOCKET_ERROR) {
		SocketClose();
		ComLog("SocketOpen.connect: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}

	ComLog("SocketOpen.End\n");

	return TRUE;
}

void CCom::SocketClose(void)
{
	ComLog("SocketClose.Begin\n");

#ifdef _WIN32
	closesocket(Socket);
#else
	shutdown(Socket, 2);
	close(Socket);
#endif

	Socket = INVALID_SOCKET;

	ComLog("SocketClose.End\n");
}


int CCom::Read(void *pBuf, int iLength, unsigned long *pdwReadLen)
{
	int rc = FALSE;

	char *DataBuf = (char *) pBuf;
	int   DataPos = 0;
	unsigned long DataLen = 0;

	if (Typ == T_S2_TCP_IP && Socket != INVALID_SOCKET) {

		for (*pdwReadLen = 0;;) {

			if ((rc = SocketRead(DataBuf + DataPos, iLength, &DataLen)) != TRUE) {
			
				break;
			}

			// Abbruch der Endlosschleife!
			if (DataLen <= 0) {

				break;
			}

			*pdwReadLen += DataLen;

			ComLog("Read.SocketRead: (%d)(%2.2d)(%2.2d)", rc, iLength, DataLen);
			if (LogIsOn && rc) {

				ComLog("(");
				for (unsigned long i = 0; i < *pdwReadLen; i++) {

					ComLog("%2.2x:", *((unsigned char *) pBuf+i));
				}
				ComLog(")");
			}
			ComLog("\n");
			
			iLength -= DataLen;
			if (iLength <= 0) {

				break;
			}

			DataPos  += DataLen;
		}

	} else if (Typ == T_S2_RS232 && hcom != INVALID_HANDLE_VALUE) {

		for (*pdwReadLen = 0;;) {

			if ((rc = SerialRead(DataBuf + DataPos, iLength, &DataLen)) != TRUE) {
			
				break;
			}

			// Abbruch der Endlosschleife!
			if (DataLen <= 0) {

				break;
			}

			*pdwReadLen += DataLen;

			ComLog("Read.SerialRead: (%10.10ld):(%d)(%2.2d)(%2.2d)", clock(), rc, iLength, DataLen);

			if (LogIsOn && rc) {

				ComLog("(");
				for (unsigned long i = 0; i < *pdwReadLen; i++) {

					ComLog("%2.2x:", *((unsigned char *) pBuf+i));
				}
				ComLog(")");
			}
			ComLog("\n");

			iLength -= DataLen;
			if (iLength <= 0) {

				break;
			}

			DataPos  += DataLen;
		}
	}

	return rc;
}

int CCom::SerialRead(char *pcInputBuf, int iInputLen, unsigned long *pdwReadLen)
{
	*pdwReadLen = 0;

	ComLog("SerialRead.Begin: (%d):(%d)\n", iInputLen, ComTimeOut);

#ifndef _WIN32
	int rc;
	struct timeval seltime;
	fd_set readfds;

	FD_ZERO((fd_set FAR *) &readfds);
	FD_SET(hcom, (fd_set FAR *) &readfds);
	seltime.tv_sec  = ComTimeOut / 1000;
	seltime.tv_usec = (ComTimeOut % 1000) * 1000;

	ComLog("SerialRead.select_0: (%d) (%d:%d)\n", rc,
		seltime.tv_sec, seltime.tv_usec);

	if ((rc = select(FD_SETSIZE,
		(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
		(struct timeval *)&seltime)) < 0) {

		ComLog("SerialRead.select: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}

	ComLog("SerialRead.select: (%d) (%d:%d)\n", rc,
		seltime.tv_sec, seltime.tv_usec);

	if (rc == 0) { // No Inputdata nach ComPortTimeOut millis

		return FALSE;
	}
#endif

#ifdef _WIN32
	if (!ReadFile(hcom, pcInputBuf, iInputLen, pdwReadLen, NULL)) {
		ComLog("SerialRead.ReadFile: failed\n");
		return FALSE;
	}
#else
	errno = 0;

	if ((rc = read(hcom, pcInputBuf, iInputLen)) < 0) {
		ComLog("SerialRead.read: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}
	*pdwReadLen = rc;
#endif

	ComLog("SerialRead.read: (%d)\n", *pdwReadLen);
	ComLog("SerialRead.End\n");
	
	return TRUE;
}

int CCom::SocketRead(char *InputBuf, int InputLen, unsigned long *ReadLen)
{
	int rc;

	struct timeval seltime;
	fd_set readfds;

	*ReadLen = 0;

	ComLog("SocketRead.Begin: (%d)\n", InputLen);

	FD_ZERO((fd_set FAR *) &readfds);
	FD_SET(Socket, (fd_set FAR *) &readfds);
	seltime.tv_sec  = ComTimeOut / 1000;
	seltime.tv_usec = (ComTimeOut % 1000) * 1000;

#ifdef _WIN32
#define CONST const
#else
#define CONST
#endif
	if ((rc = select(FD_SETSIZE,
		(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
		(CONST struct timeval FAR *)&seltime)) < 0) {
		ComLog("SocketRead.select: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}

	ComLog("SocketRead.select: (%d)\n", rc);

	if (rc == 0) { // No Inputdata nach ComPortTimeOut millis

		return FALSE;
	}

	if ((rc = recv(Socket, InputBuf, InputLen, 0)) == SOCKET_ERROR) {
		ComLog("SocketRead.recv: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}

	*ReadLen = rc;

	ComLog("SocketRead.End\n");
	
	return TRUE;
}


int CCom::Write(void *pBuf, int iLength, unsigned long *pdwWriteLen)
{
	int rc = FALSE;

	const char *DataBuf = (const char *) pBuf;

	if ((Typ == T_S2_TCP_IP) && (Socket != INVALID_SOCKET)) {
		
		*pdwWriteLen = 0;

		rc = SocketWrite(DataBuf, iLength, pdwWriteLen);

		ComLog("Write.SocketWrite: (%d)(%2.2d)(%2.2d)", rc, iLength, *pdwWriteLen);
		if (LogIsOn && rc) {

			ComLog("(");
			for (unsigned long i = 0; i < *pdwWriteLen; i++) {

				ComLog("%2.2x:", *((unsigned char *) pBuf + i));
			}
			ComLog(")");
		}
		ComLog("\n");

	} else if ((Typ == T_S2_RS232) && (hcom != INVALID_HANDLE_VALUE)) {

		*pdwWriteLen = 0;

		rc = SerialWrite(DataBuf, iLength, pdwWriteLen);

		ComLog("Write.SerialWrite: (%d)(%2.2d)(%2.2d)", rc, iLength, *pdwWriteLen);
		if (LogIsOn && rc) {

			ComLog("(");
			for (unsigned long i = 0; i < *pdwWriteLen; i++) {

				ComLog("%2.2x:", *((unsigned char *) pBuf + i));
			}
			ComLog(")");
		}
		ComLog("\n");
	}

	return rc;
}

int CCom::SerialWrite(const char *pcOutputBuf, int iOutputLen, unsigned long *pdwWriteLen)
{
	ComLog("\nSerialWrite.Begin: (%d)\n", iOutputLen);
	
#ifdef _WIN32
	if (!WriteFile(hcom, pcOutputBuf, iOutputLen, pdwWriteLen, NULL)) {
		ComLog("SerialWrite.WriteFile: failed\n");
		return FALSE;
	}
#else
	int rc;
	errno = 0;

	if ((rc = write(hcom, pcOutputBuf, iOutputLen)) < 0) {
		ComLog("SerialWrite.write: failed (rc=%d) (errno=%d)\n", rc, errno);
		return FALSE;
	}
	*pdwWriteLen = rc;
#endif

	ComLog("SerialWrite.write: (%d)\n", *pdwWriteLen);
	ComLog("SerialWrite.End\n");
	
	return TRUE;
}

int CCom::SocketWrite(const char *OutputBuf, int OutputLen, unsigned long *WriteLen)
{
	int rc;

	ComLog("SocketWrite.Begin(%d)\n", OutputLen);
	
	if ((rc = send(Socket, OutputBuf, OutputLen, 0)) == SOCKET_ERROR) {
		ComLog("SocketWrite.send: failed(%d)\n", WSAGetLastError());
		return FALSE;
	}

	*WriteLen = rc;

	ComLog("SocketWrite.End\n");
	
	return TRUE;
}

void CCom::ClearCom()
{
	unsigned char pucBuf[0x100];
	unsigned long ulReadCount = 0xFF;

	// eventuell anstehende Datenpakete verwerfen
	for (int i=0; (i < 3) && ulReadCount; i++) {

		if (!Read(pucBuf, 0xFF, &ulReadCount)) {
			
			break;
		}
	}
}


int CCom::Command(unsigned char *sbuf,int cnt,unsigned char *ebuf,int &ret, int *Err)
{
	switch (Typ) {
	case T_S2_RS232:
	case T_S2_TCP_IP:
		return CommandRsTcp(sbuf,cnt,ebuf,ret,Err);
		break;

#ifdef _WIN32 // Funk zur Zeit nur unter Windows
	case T_S2_BIM_DIREKT:
		return CommandBimDirekt(sbuf,cnt,ebuf,ret,Err);
		break;
#endif

	default:
		*Err = ERR_TYP_UNBEKANNT;
		return FALSE;
		break;
	}
}

int CCom::CommandRsTcp(unsigned char *sbuf,int cnt,unsigned char *ebuf,int &ret, int *Err)
{
	unsigned long nb;
	int Cnt;

	if (!ExeCallBack()) {*Err = ERR_ABBRUCH; return FALSE;}
	if (cnt) {
		if (!Write(sbuf,cnt,&nb)) {*Err = ERR_WRITE; return FALSE;}
		if (nb != (unsigned)cnt) {*Err = ERR_WRITE; return FALSE;}
	}

	if (ret > 0) {
		// Lnge der Antwort ist bekannt
		if (!Read(ebuf,ret,&nb)) {*Err = ERR_READ; return FALSE;}
		if (nb == 0) {*Err = ERR_NO_COM; return FALSE;}
		if (nb != (unsigned int)ret) {*Err = ERR_READ; return FALSE;}
		*Err = ERR_NO;
		TRACE("OK..\n");
		return TRUE;
	}else{
		// Lnge der Antwort ist nicht bekannt und mu aus dem Protokoll ermittelt werden
		//	Protokoll-Kopf lesen
		Cnt = 8;
		if (!Read(ebuf,Cnt,&nb)) {
			*Err = ERR_READ; 
			TRACE("ERROR_00: Read(0, %i, %i)\n", Cnt, nb);
			return FALSE;
		}
		if (nb == 0) {
			*Err = ERR_NO_COM; 
			TRACE("ERROR_01: Read(0, %i, %i)\n", Cnt, nb);
			return FALSE;
		}
		if (cnt == 0 && nb > 0 && nb < 8) {
			// Noch nicht alle Byte gelesen -> versuchen weitere Bytes zu lesen
			if (!Read(&ebuf[nb],Cnt-nb,&nb)) {
				*Err = ERR_READ; 
				TRACE("ERROR_02: Read(0, %i, %i)\n", Cnt-nb, nb);
				return FALSE;
			}
		}
		if (nb != (unsigned int)Cnt) {
			*Err = ERR_READ; 
			TRACE("ERROR_03: Read(0, %i, %i)\n", Cnt, nb);
			return FALSE;
		}
		//  Lnge des Paketes aus Paketkopf ermitteln
		Cnt = BinToUShort(&ebuf[V_O_LAENGE]);
		// Lnge auf plausibilitt prfen
		if (Cnt < 15 || Cnt > V_MAX_PAKET) {
			*Err = ERR_FORMAT; 
			TRACE("ERROR_04: %i\n", Cnt);
			return FALSE;
		}
		// Die restlichen Bytes lesen
		if (!Read(&ebuf[8],Cnt-8,&nb)) {
			*Err = ERR_READ; 
			TRACE("ERROR_05: Read(0, %i, %i)\n", Cnt-8, nb);
			return FALSE;
		}
		if (nb == 0) {*Err = ERR_NO_COM; return FALSE;}
		if (nb != (unsigned int)(Cnt-8)) {
			*Err = ERR_READ; 
			TRACE("ERROR_06: %i != %i\n", nb, Cnt-8);
			return FALSE;
		}
		ret = Cnt;
		*Err = ERR_NO;
		return TRUE;
	}
}

#ifdef _WIN32
int CCom::CommandBimDirekt(unsigned char *sbuf,int cnt,unsigned char *ebuf,int &ret, int *Err)
{
	DWORD nb;
	int rcnt=0,len,FunkSCnt,End,PCnt,GCnt,Status,ErrorZeichen,i,PreampCnt;
	unsigned char fbuf[MAX_FUNK_BUFFER];

	// Empfangsbuffer lschen (Zeichen die durch Rauschen im Empfnger endstehen)
	SetComTimeout(10);
	i = 0;
	do {
		if (!ReadFile(hcom,(LPVOID)fbuf,(DWORD)40,&nb,NULL)) return FALSE;
		i += nb;
	}while (nb >= 40 && i < 1000);
	//TRACE ("Anzahl Z. vor Kom %d\n",i);
	SetComTimeout(COM_TIMEOUT);
	// Sendebuffer aufbereiten
	FunkSCnt = MkTxdBufferFunk(sbuf,cnt,fbuf);
	// Sender Funkmodul einschalten und Anfrage senden
	//TRACE("Sendeversuch\n");
	//FunkTxOn();
	if (!WriteFile(hcom,(LPVOID)fbuf,(DWORD)FunkSCnt,&nb,NULL)) return FALSE;
	if (nb != (unsigned)FunkSCnt) return FALSE;
	// Warten bis buffer gesendet wurde + 20ms reserve; dann Funk-Sender abschalten
	//Sleep(((FunkSCnt*100000)/FUNK_BAUD)/20 + 2);
	//FunkRxOn();
	// warten auf Sequenz: Preamble -> Startzeichen
	End = FALSE; PCnt = 0; GCnt = 0;
	Status = LLM_WARTE_PREAMBLE;
	PreampCnt = 0;
	while (!End) {
		if (!ReadFile(hcom,(LPVOID)fbuf,(DWORD)1,&nb,NULL)) return FALSE;
		if (nb == 0) {
			TRACE ("Null Zeichen\n");
			*Err = ERR_NO_COM;
			return FALSE; 
		}
		GCnt++;
		if (Status == LLM_WARTE_PREAMBLE) {
			if (fbuf[0] == ZEICHEN_PREAMBLE) {
				PreampCnt++;
				if (PreampCnt > 7) Status++;
			}else{
				//TRACE ("Preamble kleiner 5\n");			
				PreampCnt=0;
			}
		}else if (Status == LLM_PREAMBLE_W1) {
			//TRACE ("Preamble W2\n");			
			if (fbuf[0] == ZEICHEN_PREAMBLE) {
				//TRACE ("Preamble W2 OK\n");			
				Status++;ErrorZeichen = 0;
			}else{
				//TRACE ("Preamble W2 Fehler\n");			
				Status = LLM_WARTE_PREAMBLE; // Es mssen min. Zeichen Z. hintereinander kommen
			}
		}else if (Status == LLM_WARTE_START) {
			//TRACE ("Warte Start %d\n",GCnt);			
			if (fbuf[0] == ZEICHEN_START) {
				//TRACE ("Start Zeichen%d\n",GCnt);			
				Status++;	// ab jetzt kommen Nutzdaten
				End = TRUE;
			}else if (fbuf[0] == ZEICHEN_PREAMBLE) {
				// weiter auf Startzeichen warten
			}else{
				//TRACE ("Error %d\n",ErrorZeichen);			
				ErrorZeichen++;
				if (ErrorZeichen > ANZ_ERRORZEICHEN) {
					// in der Preamble sind nur wenige falsche Zeichen erlaubt
					ErrorZeichen = 0;
					Status = LLM_WARTE_PREAMBLE;
				}
			}
		}
		if (GCnt > MAX_WARTEN_START) {
			TRACE ("Timeout GCnt\n");
			*Err = ERR_NO_COM;
			return FALSE;	// Timeout fr warten auf Startzeichen abgelaufen
		}
	}
	//TRACE("Nutzdaten GCnt %d\n",GCnt);
	// die ersten 8 Byte Nutzdaten Empfangen
	if (!ReadFile(hcom,(LPVOID)fbuf,(DWORD)8*2,&nb,NULL)) return FALSE;
	if (nb < 8*2) {
		TRACE("Command: 0 Byte Empfangen\n");
		*Err = ERR_NO_COM;
		return FALSE;
	}
	rcnt = nb/2;
	for (i=0;i<8;i++) ebuf[i] = diphase_to_ASCII(&fbuf[i*2]);
	// Solllnge Paket aus Kopf ermitteln
	len = BinToUShort(&ebuf[V_O_LAENGE]);
	if (len < 4) {
		// Paketlnge auserhalb Wertebereich
		TRACE("Command: len < 4 Byte %d\n",nb);
		*Err = ERR_READ;
		return FALSE;
	}
	if (len > 400) {
		// Paketlnge auserhalb Wertebereich
		TRACE("Command: len > 400 Byte %d\n",nb);
		*Err = ERR_READ;
		return FALSE;
	}
	if (rcnt == len) {
		ret = len;
		TRACE ("OK rcnt == len\n");
		*Err = ERR_NO;
		return TRUE;
	}
	// die restlichen Byte Empfangen
	if (!ReadFile(hcom,(LPVOID)fbuf,(DWORD)(len-rcnt)*2,&nb,NULL)) {
		TRACE("Command: restliche Zeichen nicht empfangen\n");
		*Err = ERR_READ;
		return FALSE;
	}
	for (i=0;i<len-rcnt;i++) ebuf[i+rcnt] = diphase_to_ASCII(&fbuf[i*2]);
	if (len == 7) { // Debug-Ausgabe
		//TRACE("Command: Befehl:%02X Fehler:%02X\n",ebuf[3],ebuf[4]);
	}else{
		//TRACE("Command: OK Befehl:%02X\n",ebuf[3]);
	}
	if ((int)nb+rcnt >= len) {
		ret = len;
		TRACE("Command: OK\n");
		*Err = ERR_NO;
		return TRUE;
	}
	TRACE("Command: Anzahl Zeichen Falsch\n");
	*Err = ERR_NO_COM;
	return FALSE;
}
#endif


int CCom::ReadMemPaket(unsigned short AdEmpfaenger, unsigned long Add, int &Cnt, unsigned char *Buf, int *Err)
// Ein Paket lesen Anwendungsschicht
{
	int Versuch,SCnt,ECnt,i;

	// Sendebuffer aufbereiten
	//	Adressen
	UShortToBin(AdEmpfaenger,SBuf);	// Adresse Empfnger
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S2]);	// Adresse Sender
	// Paketlnge
	SCnt = V_KOPF+1+1+1+2+1+sizeof(Add)+2+2;	// Sendepaket
	//ECnt = V_KOPF+1+1+1+2+1+sizeof(Add)+Cnt+2;	// Empfangspaket
	ECnt = -1;
	UShortToBin(SCnt,&SBuf[V_O_LAENGE]);
	SBuf[V_O_B_S2] = B_S2_NO;
	SBuf[V_O_PAKET_NR] = PaketNr;
	SBuf[V_O_TIME_TO_LIVE] = TimeToLive;
	SBuf[V_O_B_S3] = B_S3_NO;
	SBuf[V_O_ANZ_BLOECKE] = 1;
	SBuf[V_O_AKT_BLOCK] = 1;
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S4]);	// Adresse Sender
	SBuf[V_O_B_S4] = 4 | BB_S4_ANFRAGE;//B_S4_LESEN|BB_S4_ANFRAGE;		// Anfrage lesen
	ULongToBin(Add,&SBuf[V_O_ADD]);
	UShortToBin((unsigned short)Cnt,&SBuf[V_O_DATEN]);	// Anz. gewnschte Byte
	MkPaketCrc(SBuf,SCnt-2);
	// Daten lesen incl. Wiederholungen
	for (Versuch=0;Versuch < A_VERSUCHE; Versuch++) {
		// Im Fehlerfall zur Sicherheit groe Anzahl Zeichen lesen um Empfangsbuffer
		// des Betriebssystem zu lehren
		if (Versuch >= 2) Command(SBuf,0,EBuf,i=V_MAX_PAKET,Err);
		// Paket senden und Antwort abholen
		if (!Command(SBuf,SCnt,EBuf,ECnt,Err)) {
			continue;
		}
		// Antwort auswerten
		//	CRC prfen
		if (!CheckPaketCrc(EBuf,ECnt)) { 
			*Err = ERR_CRC;
			continue;
		}
		//	Empfnger-/Absender-Adresse prfen
		if (BinToUShort(EBuf) != AdAbsender || 
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S2]) != AdEmpfaenger ||
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S4]) != AdEmpfaenger) {
			*Err = ERR_ADD_SE;
			continue;
		}
		//	Lngenfeld prfen
		if (BinToUShort(&EBuf[V_O_LAENGE]) != ECnt) {
			*Err = ERR_FORMAT;
			continue;
		}
		//	Befehl Schicht 4 prfen
		if (EBuf[V_O_B_S4] != (B_S4_LESEN|BB_S4_ANTWORT)) {
			*Err = ERR_BEFEHL_S4;
			continue;
		}
		// Anzahl der gelesenen Datenbyte berechnen
		Cnt = ECnt - (V_KOPF+1+1+1+2+1+sizeof(Add)+2);
		// Daten in Buffer der Applikation kopieren
		for(i=0;i<Cnt;i++) Buf[i] = EBuf[i+V_O_DATEN];
		*Err = Cnt;
		return TRUE; // Daten erfolgreich gelesen
	}
	return FALSE;
}

int CCom::ReadMem(unsigned short AdEmpfaenger, unsigned long Add, int Cnt, unsigned char *Buf, int *Err)
// beliebigen Speicherbereich aus dem Gert lesen (Anwendungsschicht)
//  AdEmpfaenger -> Adresse (Gertenummer) des Empfngers
//	Add -> Adresse im Speicher des Gertes ab der gelesen wird
//	Cnt	-> Anzahl zu lesende Byte
//	Buf	-> Buffer fr gelesenen Daten
//	Err -> Bei einem Fehler -> Fehlernummer
//  return -> 1 -> lesen erfolgreich
//			  0 -> Fehler -> Fehlernummer in Var. Err	
{
	int RestCnt,Offset,ac,sc,i;

	// Schnittstelle ffnen
	if (!ComOpen()) {*Err=ERR_COM_OPEN; return FALSE;}
	// Gesammtmenge in Pakete aufteilen und lesen
	RestCnt = Cnt;
	Offset = 0;
	for (i=0;RestCnt > 0;i++) {
		if (RestCnt > A_MAX_DATEN) ac = A_MAX_DATEN; else ac = RestCnt;
		sc = ac;
		if (!ReadMemPaket(AdEmpfaenger,Add,ac,&Buf[Offset],Err)) {
			return FALSE;
		}
		RestCnt -= ac;
		Add += ac;
		Offset += ac;
		CBFunction();
		if (ac != sc) RestCnt = 0;	// Ende des Speichers im Gert erreicht
	}
	ComClose();
	*Err = Offset;	// Anzahl der gelesenen Byte
	return TRUE;
}

int CCom::WriteMemPaket(unsigned short AdEmpfaenger, unsigned long Add, int Cnt, unsigned char *Buf, int *Err)
// Ein Paket lesen Anwendungsschicht
{
	int Versuch,SCnt,ECnt,i;
	unsigned short SDatCnt;

	// Sendebuffer aufbereiten
	//	Adressen
	UShortToBin(AdEmpfaenger,SBuf);	// Adresse Empfnger
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S2]);	// Adresse Sender
	// Paketlnge
	SCnt = V_KOPF+1+1+1+2+1+sizeof(Add)+Cnt+2;	// Sendepaket
	ECnt = V_KOPF+1+1+1+2+1+sizeof(Add)+2+2;	// Empfangspaket
	UShortToBin(SCnt,&SBuf[V_O_LAENGE]);
	SBuf[V_O_B_S2] = B_S2_NO;
	SBuf[V_O_PAKET_NR] = PaketNr;
	SBuf[V_O_TIME_TO_LIVE] = TimeToLive;
	SBuf[V_O_B_S3] = B_S3_NO;
	SBuf[V_O_ANZ_BLOECKE] = 1;
	SBuf[V_O_AKT_BLOCK] = 1;
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S4]);	// Adresse Sender
	SBuf[V_O_B_S4] = B_S4_SCHREIBEN|BB_S4_ANFRAGE;		// Anfrage lesen
	ULongToBin(Add,&SBuf[V_O_ADD]);
	for (i=0;i<Cnt;i++) {
		SBuf[i+V_O_DATEN] = Buf[i];	// Daten in Sendebuffer kopieren
	}
	MkPaketCrc(SBuf,SCnt-2);
	// Daten lesen incl. Wiederholungen
	for (Versuch=0;Versuch < A_VERSUCHE; Versuch++) {
		// Im Fehlerfall zur Sicherheit groe Anzahl Zeichen lesen um Empfangsbuffer
		// des Betriebssystem zu lehren
		if (Versuch >= 2) Command(SBuf,0,EBuf,i=V_MAX_PAKET,Err);
		// Paket senden und Antwort abholen
		if (!Command(SBuf,SCnt,EBuf,ECnt,Err)) {
			continue;
		}
		// Antwort auswerten
		//	CRC prfen
		if (!CheckPaketCrc(EBuf,ECnt)) { 
			*Err = ERR_CRC;
			continue;
		}
		//	Empfnger-/Absender-Adresse prfen
		if (BinToUShort(EBuf) != AdAbsender || 
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S2]) != AdEmpfaenger ||
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S4]) != AdEmpfaenger) {
			*Err = ERR_ADD_SE;
			continue;
		}
		//	Lngenfeld prfen
		if (BinToUShort(&EBuf[V_O_LAENGE]) != ECnt) {
			*Err = ERR_FORMAT;
			continue;
		}
		//	Befehl Schicht 4 prfen
		if (EBuf[V_O_B_S4] != (B_S4_SCHREIBEN|BB_S4_ANTWORT)) {
			*Err = ERR_BEFEHL_S4;
			continue;
		}
		//  Anzahl geschriebene Daten berprfen
		SDatCnt = BinToUShort(&EBuf[V_O_DATEN]);
		*Err = SDatCnt;	// Tatschlich geschriebene Anzahl Byte
		return TRUE; // Daten erfolgreich gelesen
	}
	return FALSE;
}

int CCom::WriteMem(unsigned short AdEmpfaenger, unsigned long Add, int Cnt, unsigned char *Buf, int *Err)
// beliebigen Speicherbereich in das Gert schreiben (Anwendungsschicht)
//  AdEmpfaenger -> Adresse (Gertenummer) des Empfngers
//	Add -> Adresse im Speicher des Gertes
//	Cnt	-> Anzahl zu schreibende Byte
//	Buf	-> Buffer mit Daten
//	Err -> Bei einem Fehler -> Fehlernummer
//		   Bei einem erfolgreichen Schreibvorgang -> Anzahl geschriebene Byte (kann kleiner sein als Parameter "Cnt")
//  return -> 1 -> schreiben erfolgreich -> Anzahl Byte in Err
//			  0 -> Fehler -> Fehlernummer in Var. Err	
{
	int RestCnt,Offset,ac,i;

	// Schnittstelle ffnen
	if (!ComOpen()) {*Err=ERR_COM_OPEN; return FALSE;}
	// Gesammtmenge in Pakete aufteilen und schreiben
	RestCnt = Cnt;
	Offset = 0;
	for (i=0;RestCnt > 0;i++) {
		if (RestCnt > A_MAX_DATEN) ac = A_MAX_DATEN; else ac = RestCnt;
		if (!WriteMemPaket(AdEmpfaenger,Add,ac,&Buf[Offset],Err)) {
			return FALSE;
		}
		if (*Err != ac) {
			RestCnt = 0;
			ac = *Err;
		}else{
			RestCnt -= ac;
			Add += ac;
		}
		Offset += ac;
		CBFunction(); // Call-Back-Function
	}
	ComClose();
	*Err = Offset;
	return TRUE;
}

int CCom::OldMB(unsigned short AdEmpfaenger,unsigned char *buf, int *Len, int *Err)
// Mod-Bus Protokoll (AEIII+) in neues Protokoll verpacken, senden und Antwort zurckgeben
//  AdEmpfaenger -> Adresse (Gertenummer) des Empfngers
//	buf	-> Buffer fr Mod-Bus-Paket senden und empfangen
//	Len	-> Anzahl zu sende Byte. Nach dem Aufruf -> Anzahl Empfangene Byte
//	Err -> Bei einem Fehler -> Fehlernummer
//  return -> 1 -> senden/empfangen erfolgreich
//			  0 -> Fehler -> Fehlernummer in Var. Err	
{
	int Versuch,SCnt,ECnt,i;

	// Sendebuffer aufbereiten
	//	Adressen
	UShortToBin(AdEmpfaenger,SBuf);	// Adresse Empfnger
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S2]);	// Adresse Sender
	// Paketlnge
	SCnt = V_KOPF+1+1+1+2+1+*Len+2;	// Sendepaket
	ECnt = -1;	// Empfangspaket
	UShortToBin(SCnt,&SBuf[V_O_LAENGE]);
	SBuf[V_O_B_S2] = B_S2_NO;
	SBuf[V_O_PAKET_NR] = PaketNr;
	SBuf[V_O_TIME_TO_LIVE] = TimeToLive;
	SBuf[V_O_B_S3] = B_S3_NO;
	SBuf[V_O_ANZ_BLOECKE] = 1;
	SBuf[V_O_AKT_BLOCK] = 1;
	UShortToBin(AdAbsender,&SBuf[V_O_AD_ABSENDER_S4]);	// Adresse Sender
	SBuf[V_O_B_S4] = 0;//B_S4_LESEN|BB_S4_ANFRAGE;		// Anfrage lesen
	//ULongToBin(Add,&SBuf[V_O_ADD]);
	for (i=0;i<*Len;i++) {
		SBuf[i+V_O_ADD] = buf[i];	// Daten in Sendebuffer kopieren
		// Bei OldMB entfllt die Daten-Adresse.
		// Daher sind die 4 fr die Adresse reservierten Byte schon Daten
	}
	MkPaketCrc(SBuf,SCnt-2);
	// Daten lesen incl. Wiederholungen
	for (Versuch=0;Versuch < A_VERSUCHE_OMB; Versuch++) {
		// Im Fehlerfall zur Sicherheit groe Anzahl Zeichen lesen um Empfangsbuffer
		// des Betriebssystem zu lehren
		if (Versuch >= 2) Command(SBuf,0,EBuf,i=V_MAX_PAKET,Err);
		// Paket senden und Antwort abholen
		if (!Command(SBuf,SCnt,EBuf,ECnt,Err)) {
			continue;
		}
		// Antwort auswerten
		//	CRC prfen
		if (!CheckPaketCrc(EBuf,ECnt)) { 
			*Err = ERR_CRC;
			continue;
		}
		//	Empfnger-/Absender-Adresse prfen
		if (BinToUShort(EBuf) != AdAbsender || 
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S2]) != AdEmpfaenger ||
			  BinToUShort(&EBuf[V_O_AD_ABSENDER_S4]) != AdEmpfaenger) {
			*Err = ERR_ADD_SE;
			continue;
		}
		//	Lngenfeld prfen
		if (BinToUShort(&EBuf[V_O_LAENGE]) != ECnt) {
			*Err = ERR_FORMAT;
			continue;
		}
		//	Befehl Schicht 4 prfen
		if (EBuf[V_O_B_S4] != (B_S4_LESEN|BB_S4_ANTWORT)) {
			*Err = ERR_BEFEHL_S4;
			continue;
		}
		// Anzahl der gelesenen Datenbyte berechnen
		*Len = ECnt - (V_KOPF+1+1+1+2+1+2);
		// Daten in Buffer der Applikation kopieren
		for(i=0;i<*Len;i++) buf[i] = EBuf[i+V_O_ADD];
		*Err = ERR_NO;
		return TRUE; // Daten erfolgreich gelesen
	}
	return FALSE;
}


unsigned short CCom::BinToUShort(unsigned char *Buf)
{
	unsigned short Val;

	Val = Buf[0];
	Val <<= 8;
	Val += Buf[1];
	return Val;
}

unsigned long CCom::BinToULong(unsigned char *Buf)
{
	unsigned long Val;

	Val = Buf[0];
	Val <<= 8;
	Val += Buf[1];
	Val <<= 8;
	Val += Buf[2];
	Val <<= 8;
	Val += Buf[3];
	return Val;
}

void CCom::UShortToBin(unsigned short Val,unsigned char *Buf)
// Konvertierung MSB First
{
	Buf[0] = (unsigned char)(Val >> 8);
	Buf[1] = (unsigned char)(Val & 0xFF);
}

void CCom::ULongToBin(unsigned long Val,unsigned char *Buf)
// Konvertierung MSB First
{
	Buf[0] = (unsigned char)((Val >> 24)&0xFF);
	Buf[1] = (unsigned char)((Val >> 16)&0xFF);
	Buf[2] = (unsigned char)((Val >> 8)&0xFF);
	Buf[3] = (unsigned char)(Val & 0xFF);
}

void CCom::MkPaketCrc(unsigned char *buf,int cnt)	
// CRC berechnen und an das Ende des Buffers anhngen
{
	int i,j;
	unsigned short int crc;
  
	crc = 0xFFFF;
    for (i=0;i<cnt;i++) {
        crc = crc ^ buf[i];
        for (j=0;j<8;j++) {
            if (crc & 0x1) {crc >>= 1; crc = crc^0xA001;} else crc >>= 1;
        }
    }
    buf[cnt] = (unsigned char)(crc&0xFF);
    buf[cnt+1] = (unsigned char)(crc>>8);
}

int CCom::CheckPaketCrc(unsigned char *buf, int cnt)
{
    unsigned short int crc;
    int i,j;

    crc = 0xFFFF;
    for (i=0;i<cnt-2;i++) {
        crc = crc ^ buf[i];
        for (j=0;j<8;j++) {
            if (crc & 0x1) {crc >>= 1; crc = crc^0xA001;} else crc >>= 1;
        }
    }
    if (buf[cnt-2]!=(unsigned char)(crc&0xFF)) return FALSE;
    if (buf[cnt-1]!=(unsigned char)(crc>>8)) return FALSE;
    return TRUE;
}

void CCom::SetCallBack(int (*pCBFunction)(void))
// Pointer auf Call Back Funktion bekanntgeben
// Wenn ein Pinter vorhanden ist, wird whrend einer lngeren Kommunikation 
// regelmsig die Call-Back-Funktion aufgerufen
// Wenn die Call-B.-F. TRUE liefert wird die Kommunikation fortgesetzt
// Bei FALSE wird die aktuelle Kommunikation abgebrochen
// DieCall-Back-Funktion kann auch fr eine Fortschrittsanzeige 
// in der Applikation verwendet werden.
{
	CBFunction = pCBFunction;
}

int CCom::ExeCallBack()
{
	if (CBFunction != NULL) return CBFunction(); else return TRUE;
}

#ifdef _WIN32
void CCom::ASCII_to_diphase(unsigned char Zeichen, unsigned char *Pointer_String)
   {
   /*
   Umwandlung eines ASCII-Zeichen in einen Conditioned Diphase-String (2 Byte),
   der ber einen Zeiger bergeben wird.
   */
   unsigned char i;        /* Laufvariable fr Umwandlung */
   
   union U_intbyte Zeichen_diphase;
   Zeichen_diphase.WortWert = 0;
   for (i = 0; i < 8; i++               /* Schleife ber ein Zeichen */)
      {
      /* Die beiden neuen Bits um 2 Bit nach rechts verschieben */
         Zeichen_diphase.WortWert = Zeichen_diphase.WortWert >> 2;
      if ((Zeichen & 0x01) == 1         /* aktuelles Bit = 1 ? */)
         {
         /* 1 wird in 11 bzw. 00 umgewandelt */
         if (/* War letztes Bit (also Bit 13) 1 und nicht Startbit nach Bit 4 ? */
                ((Zeichen_diphase.WortWert & 0x2000) != 0) && (i != 4))
            {
            /* Wenn vorher 1, dann als nchstes 0, hier also 00 */
            /* Fr 00 muss nichts geschrieben werden, da oben alles 0 war. */
            }
         else
            {
            /* Wenn vorher 0, dann als nchstes 1, hier also 11 */
            /* Die beiden neuen Bits von oben in das Ergebnis schreiben */
               Zeichen_diphase.WortWert |= 0xC000;
            }
         }
      else
         {
         /* 0 wird in 10 bzw. 01 umgewandelt */
         if (/* War letztes Bit (also Bit 13) 1 und nicht Startbit nach Bit 4 ? */
                ((Zeichen_diphase.WortWert & 0x2000) != 0) && (i != 4))
            {
            /* Wenn vorher 1, dann als nchstes 0, hier also 10 */
            /* Die beiden neuen Bits von oben in das Ergebnis schreiben */
               Zeichen_diphase.WortWert |= 0x8000;
            }
         else
            {
            /* Wenn vorher 0, dann als nchstes 1, hier also 01 */
            /* Die beiden neuen Bits von oben in das Ergebnis schreiben */
               Zeichen_diphase.WortWert |= 0x4000;
            }
         }
      /* Im Ursprungszeichen um 1 Bit nach rechts verschieben */
         Zeichen = Zeichen >> 1;
      }
   /* Ergebnis in String zurckgeben, Bits 0..7 zuerst, dann Bits 8..15 */
      *Pointer_String++ = Zeichen_diphase.ByteWert[0];
      *Pointer_String = Zeichen_diphase.ByteWert[1];
     
     i = Zeichen_diphase.ByteWert[0];
     i = Zeichen_diphase.ByteWert[1];
   }
#endif

#ifdef _WIN32
unsigned char CCom::diphase_to_ASCII (unsigned char *Pointer_String)
   {
   /*
   Umwandlung eines Conditioned Diphase-String (2 Byte) in ein ASCII-Zeichen
   Als Rckgabewert wird das umgewandelte Zeichen zurckgegeben.
   */
   unsigned char i;        /* Laufvariable fr Umwandlung */
   unsigned char Zeichen;
   
   union U_intbyte Zeichen_diphase;
   Zeichen = 0;
   /* Ergebnis in String zurckgeben, Bits 0..7 zuerst, dann Bits 8..15 */
      Zeichen_diphase.ByteWert[0] = *Pointer_String++;
      Zeichen_diphase.ByteWert[1] = *Pointer_String;
   for (i = 0; i < 8; i++               /* Schleife ber ein Zeichen */)
      {
      /* Das neue Bit um 1 Bit nach rechts verschieben */
         Zeichen = Zeichen >> 1;
      if (/* Bits = 01 oder 10 ? */
             ((Zeichen_diphase.WortWert & 0x03) == 1) || 
             ((Zeichen_diphase.WortWert & 0x03) == 2))
         {
         /* 10 und 01 werden 0 umgewandelt */
         /* Fr 0 muss nichts getan werden, da oben schon zugewiesen */
         }
      else
         {
         /* 00 und 11 werden in 1 umgewandelt */
            Zeichen |= 0x80;
         }
      /* Im Ursprungszeichen um 1 Bit nach rechts verschieben */
         Zeichen_diphase.WortWert = Zeichen_diphase.WortWert >> 2;
      }
   return Zeichen;
   }
#endif

#ifdef _WIN32
int CCom::MkTxdBufferFunk(unsigned char *sbuf,int cnt,unsigned char *FunkBuf)
// Buffer fr Funk aufbereiten
{
	int i,j;

	// Preamble aufbereiten
	for (i=0;i<LEN_PREAMBLE;i++) FunkBuf[i] = ZEICHEN_PREAMBLE;
	// Sync.- und Startbyte aufbereiten
	FunkBuf[i++] = ZEICHEN_SYNCHRO; FunkBuf[i++] = ZEICHEN_SYNCHRO;
	FunkBuf[i++] = ZEICHEN_START;
	ASSERT ((i+cnt*2) < MAX_FUNK_BUFFER); // Bufferberlauf testen
	// Nutzdaten in diphase wandeln
	for (j=0;j<cnt;j++) {
		ASCII_to_diphase(sbuf[j],&FunkBuf[i+j*2]);
	}
	return i+j*2;
}
#endif

